#include "mglog.h"
#include "getopt.h"
#include <sys/statfs.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>

extern char g_platform_choice[16];
extern int mglog_ignore_exx00u_ap;
extern char modem_name_para[32];

struct mgl_usb_device_info s_usb_device_info[MAX_USB_DEV];

struct usbfs_getdriver
{
    unsigned int interface;
    char driver[256];
};

struct usbfs_ioctl
{
    int ifno;       /* interface 0..N ; negative numbers reserved */
    int ioctl_code; /* MUST encode size + direction of data so the
			 * macros in <asm/ioctl.h> give correct values */
    void *data;     /* param buffer (in, or out) */
};

#define IOCTL_USBFS_DISCONNECT _IO('U', 22)
#define IOCTL_USBFS_CONNECT _IO('U', 23)

static int usbfs_is_kernel_driver_alive(int fd, int ifnum)
{
    struct usbfs_getdriver getdrv;
    getdrv.interface = ifnum;
    if (ioctl(fd, USBDEVFS_GETDRIVER, &getdrv) < 0)
    {
        if (errno != ENODATA)
            mglog_dbg("%s ioctl USBDEVFS_GETDRIVER on interface %d failed, kernel driver may be inactive\n", __func__, ifnum);
        return 0;
    }
    mglog_dbg("%s find interface %d has match the driver %s\n", __func__, ifnum, getdrv.driver);
    return 1;
}

static void usbfs_detach_kernel_driver(int fd, int ifnum)
{
    struct usbfs_ioctl operate;
    operate.data = NULL;
    operate.ifno = ifnum;
    operate.ioctl_code = IOCTL_USBFS_DISCONNECT;
    if (ioctl(fd, USBDEVFS_IOCTL, &operate) < 0)
        mglog_dbg("%s detach kernel driver failed\n", __func__);
    else
        mglog_dbg("%s detach kernel driver success\n", __func__);
}

static int get_value_from_file(const char *fname, int base)
{
    char buff[64] = {'\0'};

    int fd = open(fname, O_RDONLY);
    if (fd <= 0)
    {
        if (errno != ENOENT)
            mglog_dbg("Fail to open %s,  errno: %d (%s)\n", fname, errno, strerror(errno));
        return -1;
    }
    if (read(fd, buff, sizeof(buff)) == -1) {
    }
    close(fd);
    return strtoul(buff, NULL, base);
}

static int strStartsWith(const char *line, const char *prefix)
{
    for ( ; *line != '\0' && *prefix != '\0' ; line++, prefix++) {
        if (*line != *prefix) {
            return 0;
        }
    }

    return *prefix == '\0';
}

static const char *get_value_from_uevent(const char *uevent,  const char *key) {
    FILE *fp;
    static char line[256];

    fp = fopen(uevent, "r");
    if (fp == NULL) {
        mglog_dbg("fail to fopen %s, errno: %d (%s)\n", uevent, errno, strerror(errno));
        return "-1";
    }

    //dbg_time("%s\n", uevent);
    while (fgets(line, sizeof(line), fp)) {
        if (line[strlen(line) - 1] == '\n' || line[strlen(line) - 1] == '\r') {
            line[strlen(line) - 1] = '\0';
        }

        //dbg_time("%s\n", line);
        if (strStartsWith(line, key) && line[strlen(key)] == '=') {
            fclose(fp);
            return &line[strlen(key)+1];
        }
    }

    fclose(fp);

    return "-1";
}

static void mgl_get_dm_major_and_minor(const char *devname, struct mgl_usb_device_info *mgl_dev)
{
    char devpath[256];

    if(devname[0] == '\0')
        return;

    snprintf(devpath, sizeof(devpath), "/sys/class/tty/%.27s/uevent", devname + strlen("/dev/"));
    if(access(devpath, F_OK) && errno == ENOENT)
        return;

    mgl_dev->dm_major = atoi(get_value_from_uevent(devpath, "MAJOR"));
    mgl_dev->dm_minor = atoi(get_value_from_uevent(devpath, "MINOR"));
}

int mgl_match_dm_device(const char *devname, struct mgl_usb_device_info *mgl_dev)
{
    struct stat buf;

    if (!stat(devname, &buf) && (int)major(buf.st_rdev) == mgl_dev->dm_major && (int)minor(buf.st_rdev) == mgl_dev->dm_minor) {
        snprintf(mgl_dev->ttyDM, sizeof(mgl_dev->ttyDM), "%.31s", devname);
        return 0;
    }

    return 1;
}

static void mgl_find_usb_interface_info(const char *intfpath, struct mgl_usb_interface_info *pIntf)
{
    DIR *dev_dir = NULL;
    struct dirent *dev = NULL;
    char devpath[256+32];

    dev_dir = opendir(intfpath);
    if (!dev_dir) {
        mglog_dbg("fail to opendir('%s'), errno: %d (%s)\n", intfpath, errno, strerror(errno));
        return;
    }

    while (NULL != (dev = readdir(dev_dir)))
    {
        if (!strncasecmp(dev->d_name, "ep_", 3)) {
            int ep = strtoul(&dev->d_name[3], NULL, 16);
            if (ep&0x80)
                pIntf->ep_in = ep;
            else
                 pIntf->ep_out = ep;
       }
    }

    closedir(dev_dir);

    snprintf(devpath, sizeof(devpath), "%.256s/bInterfaceClass", intfpath);
    pIntf->bInterfaceClass = get_value_from_file(devpath, 16);

    snprintf(devpath, sizeof(devpath), "%.256s/bInterfaceSubClass", intfpath);
    pIntf->bInterfaceSubClass = get_value_from_file(devpath, 16);

    snprintf(devpath, sizeof(devpath), "%.256s/bInterfaceProtocol", intfpath);
    pIntf->bInterfaceProtocol = get_value_from_file(devpath, 16);

    snprintf(devpath, sizeof(devpath), "%.256s/bNumEndpoints", intfpath);
    pIntf->bNumEndpoints = get_value_from_file(devpath, 16);

#if 0
    mglog_dbg("Interface Class: %02x, SubClass: %02x, Protocol: %02x\n",
        pIntf->bInterfaceClass, pIntf->bInterfaceSubClass, pIntf->bInterfaceProtocol);
    mglog_dbg("NumEndpoints: %d, ep_in: %02x, ep_out: %02x\n", pIntf->bNumEndpoints, pIntf->ep_in, pIntf->ep_out);
#endif
}

static void mgl_find_tty_info(const char *intfpath, char *ttyport, size_t len)
{
    DIR *dev_dir = NULL;
    struct dirent *dev = NULL;
    char devpath[256];
    int wait_tty_register = 10;

    ttyport[0] = '\0';

_scan_tty:
    // find tty device
    dev_dir = opendir(intfpath);
    if (!dev_dir) {
        mglog_dbg("fail to opendir('%s'), errno: %d (%s)\n", intfpath, errno, strerror(errno));
        return;
    }

    while (NULL != (dev = readdir(dev_dir)))
    {
        if (!strncasecmp(dev->d_name, "tty", 3)) {
            snprintf(ttyport, len, "/dev/%.16s", dev->d_name);
            break;
        }
    }

    closedir(dev_dir);

    if (!strcmp(ttyport, "/dev/tty")) { //find tty not ttyUSBx or ttyACMx
        snprintf(devpath, sizeof(devpath), "%.128s/tty", intfpath);

        dev_dir = opendir(devpath);
        if (dev_dir)
        {
            while (NULL != (dev = readdir(dev_dir)))
            {
                if (!strncasecmp(dev->d_name, "tty", 3))
                {
                        snprintf(ttyport, len, "/dev/%.16s", dev->d_name);
                        break;
                }
            }
            closedir(dev_dir);
        }
    }

    if (ttyport[0] == '\0' && wait_tty_register) {
        usleep(100*1000); //maybe other files not ready
        wait_tty_register--;
        goto _scan_tty;
    }
}

int mgl_find_meig_modules(void)
{
    DIR *usb_dir = NULL;
    struct dirent *usb = NULL;
    const char *usbpath = "/sys/bus/usb/devices";
    struct mgl_usb_device_info mgl_dev;
    int modules_num = 0;

    usb_dir = opendir(usbpath);
    if (NULL == usb_dir)
        return modules_num;

    while (NULL != (usb = readdir(usb_dir)))
    {
        char devpath[256] = {'\0'};
        if (usb->d_name[0] == '.' || usb->d_name[0] == 'u')
            continue;

        memset(&mgl_dev, 0, sizeof(struct mgl_usb_device_info));

        snprintf(devpath, sizeof(devpath), "%.24s/%.16s/idVendor", usbpath, usb->d_name);
        mgl_dev.idVendor = get_value_from_file(devpath, 16);

        snprintf(devpath, sizeof(devpath), "%.24s/%.16s/idProduct", usbpath, usb->d_name);
        mgl_dev.idProduct = get_value_from_file(devpath, 16);

        if ((mgl_dev.idVendor == 0x3763 && mgl_dev.idProduct == 0x3c93))
        {
            if (!modem_name_para[0])
            {
                printf("The state grid module needs to specify the module, like -g SLM750T\n");
                return modules_num;
            }
        }

        if ((mgl_dev.idVendor == 0x05c6 && mgl_dev.idProduct == 0xf601) //SLM750/SLM730
            || (mgl_dev.idVendor == 0x05c6 && mgl_dev.idProduct == 0x9090) //
            || (mgl_dev.idVendor == 0x05c6 && mgl_dev.idProduct == 0x9215) //
            || (mgl_dev.idVendor == 0x05c6 && mgl_dev.idProduct == 0x90db) //sdx12
            || (mgl_dev.idVendor == 0x3763 && mgl_dev.idProduct == 0x3c93) //asr 
            || (mgl_dev.idVendor == 0x3c93 && mgl_dev.idProduct == 0xffff) //asr
            || (mgl_dev.idVendor == 0x2dee && mgl_dev.idProduct == 0x4d5c) //zhangqingyun add for lcm720N
            || (mgl_dev.idVendor == 0x2dee && mgl_dev.idProduct == 0x4d54) //zhangqingyun add for LCM720 dump
            || drv_is_unisoc(mgl_dev.idProduct,mgl_dev.idVendor) //SRM8X1
            || drv_is_unisoc_exx00u(mgl_dev.idProduct,mgl_dev.idVendor) //SLM730/750 SRM8X5 etc
            || drv_is_mtk(mgl_dev.idProduct,mgl_dev.idVendor) //RM500K
            || drv_is_asr(mgl_dev.idProduct, mgl_dev.idVendor) //ASR
            || drv_is_qualcomm(mgl_dev.idProduct, mgl_dev.idVendor)
            )
        {
        }
        else {
            continue;
        }

        usleep(100*1000); //maybe other files not ready

        snprintf(devpath, sizeof(devpath), "%.24s/%.16s/bNumInterfaces", usbpath, usb->d_name);
        mglog_dbg("devpath is:%s",devpath);
        mgl_dev.bNumInterfaces = get_value_from_file(devpath, 10);
        mglog_dbg("num interfaces is:%d",mgl_dev.bNumInterfaces);
        snprintf(devpath, sizeof(devpath), "%.24s/%.16s/uevent", usbpath, usb->d_name);
        mgl_dev.major = atoi(get_value_from_uevent(devpath, "MAJOR"));
        mgl_dev.minor = atoi(get_value_from_uevent(devpath, "MINOR"));
        mgl_dev.busnum = atoi(get_value_from_uevent(devpath, "BUSNUM"));
        mgl_dev.devnum = atoi(get_value_from_uevent(devpath, "DEVNUM"));
        strncpy(mgl_dev.devname, get_value_from_uevent(devpath, "DEVNAME"), sizeof(mgl_dev.devname));

        snprintf(devpath, sizeof(devpath), "%s/%.16s/bcdDevice", usbpath, usb->d_name);
        mgl_dev.bcdDevice = get_value_from_file(devpath, 10);

        memset(&mgl_dev.dm_intf, 0xff, sizeof(struct mgl_usb_interface_info));
        memset(&mgl_dev.general_intf, 0xff, sizeof(struct mgl_usb_interface_info));
        mgl_dev.general_type = -1;

        mgl_dev.dm_intf.bInterfaceNumber = 0;    //include SLM750
        if (mgl_dev.bNumInterfaces > 1) {
            if (drv_is_asr(mgl_dev.idProduct, mgl_dev.idVendor)) { //ASR
                if (!strncasecmp(modem_name_para, "SLM770A", 6))   //SLM770 
                {
                    if (mgl_dev.idVendor == 0x3c93 && mgl_dev.idProduct == 0xffff)
                        mgl_dev.dm_intf.bInterfaceNumber = 8;
                    else if (mgl_dev.idVendor == 0x3763 && mgl_dev.idProduct == 0x3c93)
                        mgl_dev.dm_intf.bInterfaceNumber = 0;
                }
                else if (!strncasecmp(modem_name_para, "SLB780", 6))   //SLM780 GW
                {
                    if (mgl_dev.idVendor == 0x3c93 && mgl_dev.idProduct == 0xffff)
                        mgl_dev.dm_intf.bInterfaceNumber = 8;
                    else if (mgl_dev.idVendor == 0x3763 && mgl_dev.idProduct == 0x3c93)
                        mgl_dev.dm_intf.bInterfaceNumber = 0;
                }
                else
                    mgl_dev.dm_intf.bInterfaceNumber = 2;
            }
            else if (drv_is_unisoc(mgl_dev.idProduct,mgl_dev.idVendor)) { //unisoc
                mglog_dbg("unisoc \n");
                if (!strncasecmp(modem_name_para, "SRM821", 6))   //FOR CHINA UNICOM
                {
                    if (mgl_dev.idVendor == 0x3c93 && mgl_dev.idProduct == 0xffff)
                    {
                        mgl_dev.dm_intf.bInterfaceNumber = 8;
                        mgl_dev.general_intf.bInterfaceNumber = 10; //log_intf
                        mgl_dev.general_type = RG500U_LOG;
                    }
                }
                else if (!strncasecmp(modem_name_para, "SRM811", 6))  //
                {
                    if (mgl_dev.idVendor == 0x3763 && mgl_dev.idProduct == 0x3c93)
                    {
                        mgl_dev.dm_intf.bInterfaceNumber = 5;       //dm_intf
                        mgl_dev.general_intf.bInterfaceNumber = 6;  //log_intf
                        mgl_dev.general_type = RG500U_LOG;
                    }
                    else if (mgl_dev.idVendor == 0x3c93 && mgl_dev.idProduct == 0xffff)
                    {
                        mgl_dev.dm_intf.bInterfaceNumber = 8;       //dm_intf
                        mgl_dev.general_intf.bInterfaceNumber = 10;  //log_intf
                        mgl_dev.general_type = RG500U_LOG;
                    }
                }
                else if (mgl_dev.idVendor == 0x2c7c && mgl_dev.idProduct == 0x0902)  //8310 SLM750D-CN
                {
                    mgl_dev.dm_intf.bInterfaceNumber = 3;       //dm_intf
                    mgl_dev.general_intf.bInterfaceNumber = 5;  //log_intf
                    mgl_dev.general_type = RG500U_LOG;
                }
                else                                           //udx710 standard RG500U/RM500U/RG200U/...
                {
                    mglog_dbg("default is:\n");
                    mgl_dev.dm_intf.bInterfaceNumber = 2;       
                    mgl_dev.general_intf.bInterfaceNumber = 5;  //log_intf//zhangqingyun add for udx710 platform 
                    mgl_dev.general_type = RG500U_LOG;
                }
            }
            else if (drv_is_unisoc_exx00u(mgl_dev.idProduct,mgl_dev.idVendor)) {  //unisoc SRM8X1 or EGx00U
                if (mgl_dev.idVendor == 0x2c7c && mgl_dev.idProduct == 0x0904)
                {
                    mgl_dev.dm_intf.bInterfaceNumber = 5;            //cplog_intf
                    mgl_dev.general_intf.bInterfaceNumber = 3;       //aplog_intf
                    mgl_dev.general_type = SLM750U_AP;
                }
                else
                {
                    mgl_dev.dm_intf.bInterfaceNumber = 5;            //cplog_intf
                    if (!mglog_ignore_exx00u_ap)
                    {
                        mgl_dev.general_intf.bInterfaceNumber = 6;   //aplog_intf
                        mgl_dev.general_type = SLM750U_AP;
                    }
                }
            }
            else if (drv_is_mtk(mgl_dev.idProduct,mgl_dev.idVendor)) {  //mtk RM500K
                mgl_dev.dm_intf.bInterfaceNumber = 9;
            }
            else if (mgl_dev.idVendor == 0x05c6 && mgl_dev.idProduct == 0x90db) {
                mgl_dev.general_intf.bInterfaceNumber = 4;         //qdss_intf
                mgl_dev.general_type = MDM_QDSS;
           }
            
           //zhangqingun add for support qualcom platform interfade 0 is the most common use as diag port
            else if (drv_is_qualcomm(mgl_dev.idProduct,mgl_dev.idVendor)){
                 mgl_dev.dm_intf.bInterfaceNumber = 0;
                 mgl_dev.general_type = MDM_QDSS;
            }
            else if (mgl_dev.idVendor == 0x2dee && mgl_dev.idProduct == 0x4d5c) {  //zhangqigyun add for lcm720N
                mgl_dev.dm_intf.bInterfaceNumber = 5;        //qdss_intf
                mgl_dev.general_type = MDM_QDSS;
                
            }
            
        }

_rescan_dm:
        snprintf(devpath, sizeof(devpath), "%.24s/%.16s/%.16s:1.%d", usbpath, usb->d_name, usb->d_name, mgl_dev.dm_intf.bInterfaceNumber);
        mglog_dbg("zqy rescan_dm devpath is:%s\n",devpath);
        if (access(devpath, F_OK))
            continue;

        mgl_find_usb_interface_info(devpath, &mgl_dev.dm_intf);
        if (mgl_dev.dm_intf.bInterfaceNumber == 0
            && mgl_dev.dm_intf.bInterfaceClass == 0x02
            && mgl_dev.dm_intf.bInterfaceSubClass == 0x0e) { //EM05-G 's interface 0 is MBIM
            mgl_dev.dm_intf.bInterfaceNumber = 3;
            goto _rescan_dm;
        }

        mgl_find_tty_info(devpath, mgl_dev.ttyDM, sizeof(mgl_dev.ttyDM));
        if (mgl_dev.ttyDM[0]) {
            mgl_get_dm_major_and_minor(mgl_dev.ttyDM, &mgl_dev);
        }

        if (mgl_dev.general_intf.bInterfaceNumber != 0xFF
            && (mgl_dev.general_type == SLM750U_AP || mgl_dev.general_type == RG500U_LOG)) {
            snprintf(devpath, sizeof(devpath), "%.24s/%.16s/%.16s:1.%d", usbpath, usb->d_name, usb->d_name, mgl_dev.general_intf.bInterfaceNumber);
            if (access(devpath, F_OK))
                continue;

            mgl_find_usb_interface_info(devpath, &mgl_dev.general_intf);
            mgl_find_tty_info(devpath, mgl_dev.ttyGENERAL, sizeof(mgl_dev.ttyGENERAL));
        }
        else if (mgl_dev.general_intf.bInterfaceNumber != 0xFF
            && (mgl_dev.general_type == MDM_QDSS)) {

            snprintf(devpath, sizeof(devpath), "%.24s/%.16s/%.16s:1.%d", usbpath, usb->d_name, usb->d_name, mgl_dev.general_intf.bInterfaceNumber);
            if (!access(devpath, F_OK)) {
                mgl_find_usb_interface_info(devpath, &mgl_dev.general_intf);
                if (mgl_dev.general_intf.bInterfaceClass != 0xff
                     ||mgl_dev.general_intf.bInterfaceSubClass != 0xff
                     || mgl_dev.general_intf.bInterfaceProtocol != 0x70
                     || mgl_dev.general_intf.bNumEndpoints != 1
                     || mgl_dev.general_intf.ep_out != 0xff
                     || mgl_dev.general_intf.ep_in == 0xff) {
                    mglog_dbg("not a vaild qdss usb interface!\n");
                    memset(&mgl_dev.general_intf, 0xff, sizeof(struct mgl_usb_interface_info));
                    mgl_dev.general_type = -1;
                }
            }else {
                memset(&mgl_dev.general_intf, 0xff, sizeof(struct mgl_usb_interface_info));
                mgl_dev.general_type = -1;
            }
        }

        snprintf(mgl_dev.usbdevice_pah, sizeof(mgl_dev.usbdevice_pah), "%.24s/%.16s", usbpath, usb->d_name);
        mglog_dbg("Find [%d] idVendor=%04x, idProduct=%04x, bNumInterfaces=%d, ttyDM=%s, ttyGENERAL=%s, busnum=%03d, dev=%03d, usbdevice_pah=%s\n",
            modules_num, mgl_dev.idVendor, mgl_dev.idProduct, mgl_dev.bNumInterfaces, mgl_dev.ttyDM,
            mgl_dev.ttyGENERAL, mgl_dev.busnum, mgl_dev.devnum, mgl_dev.usbdevice_pah);

        if (modules_num < MAX_USB_DEV)
            s_usb_device_info[modules_num++] = mgl_dev;
    }

    closedir(usb_dir);
    return modules_num;
}

int mgl_usbfs_open_interface(const struct mgl_usb_device_info *usb_dev, int intf)
{
    char devname[64];
    int dev_mknod_and_delete_after_use = 0;
    int usbfd = -1;
    int ret;

    snprintf(devname, sizeof(devname), "/dev/%s", usb_dev->devname);
    if (access(devname, F_OK) && errno == ENOENT) {
        char *p = strstr(devname+strlen("/dev/"), "/");

        while (p) {
            p[0] = '_';
            p = strstr(p, "/");
        }

#define MKDEV(__ma, __mi) (((__ma & 0xfff) << 8) | (__mi & 0xff) | ((__mi & 0xfff00) << 12))
        if (mknod(devname, S_IFCHR|0666, MKDEV(usb_dev->major, usb_dev->minor))) {
            devname[1] = 't';
            devname[2] = 'm';
            devname[3] = 'p';

            if (mknod(devname, S_IFCHR|0666, MKDEV(usb_dev->major, usb_dev->minor))) {
                mglog_dbg("Fail to mknod %s, errno : %d (%s)\n", devname, errno, strerror(errno));
            }
        }

        dev_mknod_and_delete_after_use = 1;
    }

    usbfd = open(devname, O_RDWR | O_NDELAY);
    if (dev_mknod_and_delete_after_use) {
        remove(devname);
    }

    if (usbfd == -1) {
        mglog_dbg("usbfs open %s failed, errno: %d (%s)\n", devname, errno, strerror(errno));
        return -1;
    }

    if (usbfs_is_kernel_driver_alive(usbfd, intf))
        usbfs_detach_kernel_driver(usbfd, intf);

    ret = ioctl(usbfd, USBDEVFS_CLAIMINTERFACE, &intf); // attach usbfs driver
    if (ret != 0)
    {
        mglog_dbg("ioctl USBDEVFS_CLAIMINTERFACE failed, errno = %d(%s)\n", errno, strerror(errno));
        close(usbfd);
        return -1;
    }

    return usbfd;
}

int mgl_usbfs_read(int usbfd, int ep_in, void *pbuf, unsigned len)
{
    struct usbdevfs_bulktransfer bulk;
    int n = 0;

    bulk.ep = ep_in;
    bulk.len = len;
    bulk.data = (void *)pbuf;
    bulk.timeout = 0; // keep waiting

    n = ioctl(usbfd, USBDEVFS_BULK, &bulk);
    if (n < 0) {
        mglog_dbg("%s n = %d, errno: %d (%s)\n", __func__, n, errno, strerror(errno));
    }
    else if (n == 0) {
        //zero length packet
    }

    return n;
}

int mgl_usbfs_write(int usbfd, int ep_out, const void *pbuf, unsigned len)
{
    struct usbdevfs_urb bulk;
    struct usbdevfs_urb *urb = &bulk;
    int n = 0;

    memset(urb, 0, sizeof(struct usbdevfs_urb));
    urb->type = USBDEVFS_URB_TYPE_BULK;
    urb->endpoint = ep_out;
    urb->status = -1;
    urb->buffer = (void *)pbuf;
    urb->buffer_length = len;
    urb->usercontext = urb;
    urb->flags = 0;

    n = ioctl(usbfd, USBDEVFS_SUBMITURB, urb);
    if (n < 0) {
        mglog_dbg("%s submit n = %d, errno: %d (%s)\n", __func__, n, errno, strerror(errno));
        return 0;
    }

    urb = NULL;
    n = ioctl(usbfd, USBDEVFS_REAPURB, &urb);
    if (n < 0) {
        mglog_dbg("%s reap n = %d, errno: %d (%s)\n", __func__, n, errno, strerror(errno));
        return 0;
    }

    if (urb && urb->status == 0 && urb->actual_length) {
        // mglog_dbg("urb->actual_length = %u\n", urb->actual_length);
        return urb->actual_length;
    }

    return 0;
}

